unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  StdCtrls, ExtCtrls, CheckLst;

type
  TMainForm = class(TForm)
    ControlPanel: TPanel;
    CDLabel: TLabel;
    MiniPackRadio: TRadioGroup;
    Splitter1: TSplitter;
    Panel6: TPanel;
    MessLabel: TLabel;
    MessList: TListBox;
    CoNickLabel: TLabel;
    Panel3: TPanel;
    INFEdit: TEdit;
    INFLabel: TLabel;
    OutLabel: TLabel;
    Panel5: TPanel;
    OutEdit: TEdit;
    Label1: TLabel;
    Panel1: TPanel;
    Panel2: TPanel;
    CoNickEdit: TEdit;
    PathEdit: TEdit;
    Panel4: TPanel;
    InstSrcEdit: TEdit;
    Panel7: TPanel;
    PushCheck: TCheckBox;
    GoPanel: TPanel;
    GoButton: TButton;
    procedure GoButtonClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure PushCheckClick(Sender: TObject);
  private
    { Private declarations }
  public
    procedure ClearAllClick(Sender: TObject);
    procedure SelectAllClick(Sender: TObject);
    { Public declarations }
  end;

var
  MainForm: TMainForm;

implementation

{$R *.DFM}

uses
  FileCtrl, ShellAPI, INIFiles, TintProgresses, Clipbrd;

  //Once upon a time, LocMode was this simple:
  //0 = target
  //1 = windows
  //2 = DLLs (system or system32)
  //3 = drivers in 9x (system)
  //4 = drivers (system or system32\drivers)
  //5 = inf (inf\other or inf, with special handling)

const
  WebPackPath = 'M:\Database\AIOWEB\files\packages\';
  SizerStr = 'zxcvXXXXvbnm';

type
  TMiniAction = (maPass, maPad, maFail);

function StringReplaceC(const S: String; const OldChar, NewChar: Char): String;
var
  I: Integer;

begin
  Result := S;
  for I := 1 to Length(Result) do
    if Result[I] = OldChar then
      Result[I] := NewChar
  ;
end;

procedure DumpStringToFile(S, Path: String);
var
  Fil: TFileStream;

begin
  Fil := TFileStream.Create(Path, fmCreate or fmShareDenyWrite);
  try
    Fil.Write(S[1], Length(S));
  finally
    Fil.Free;
  end;
end;

function DumpFileToString(Path: String): String;
var
  Fil: TFileStream;
  L: Integer;

begin
  Fil := TFileStream.Create(Path, fmOpenRead or fmShareDenyNone);
  try
    L := Fil.Size;
    SetString(Result, nil, L);
    Fil.Read(Result[1], L);
  finally
    Fil.Free;
  end;
end;

function DumpFileNamesToStringList(Mask: String): TStringList;
var
  Findee: TSearchRec;

begin
  Result := TStringList.Create;
  if FindFirst(Mask, faAnyFile, Findee) = 0 then repeat
    if Findee.Name[1] = '.' then Continue; //Skip virtual dirs and control files like .htaccess
    if (Findee.Attr and faDirectory) <> 0 then Continue;
    Result.Add(Findee.Name);
  until FindNext(Findee) <> 0;
  FindClose(Findee);
end;

function DumpAllSubFileNamesToStringList(BasePath: String): TStringList;
var
  Findee: TSearchRec;
  Dirs: TStringList;
  I: Integer;
  SearchPath: String;

begin
  BasePath := IncludeTrailingBackslash(BasePath);
  SearchPath := '';
  
  Result := TStringList.Create;
  Dirs := TStringList.Create;
  try
    repeat
      if FindFirst(BasePath + SearchPath + '*.*', faAnyFile, Findee) = 0 then repeat
        if Findee.Name[1] = '.' then Continue; //Skip virtual dirs and control files like .htaccess
        if (Findee.Attr and faDirectory) = 0 then
          Result.Add(SearchPath + Findee.Name)
        else
          Dirs.Add(SearchPath + Findee.Name + '\')
        ;
      until FindNext(Findee) <> 0;
      FindClose(Findee);

      I := Dirs.Count;
      if I = 0 then Break;

      Dec(I);
      SearchPath := Dirs[I];
      Dirs.Delete(I);
    until False;
  finally
    Dirs.Free;
  end;
end;

function StrToLStr(S: String): String;
var
  L: Integer;

begin
  L := Length(S);
  SetString(Result, PChar(@L), 4);
  Result := Result + S;
end;

function StrToIncLStr(S: String): String;
var
  L: Integer;

begin
  L := Length(S) + 4;
  SetString(Result, PChar(@L), 4);
  Result := Result + S;
end;

function IntAsStr(I: Integer): String;
begin
  SetString(Result, PChar(@I), 4);
end;

function RValue(S: String): String;
var
  I: Integer;

begin
  I := Pos('=', S);
  if I = 0 then
    Result := ''
  else
    Result := Copy(S, I + 1, MAXINT)
  ;
end;

function TimeStr: String;
var
  ST: TSystemTime;

begin
  GetLocalTime(ST);
  Result := Format('%d:%.2d:%.2d.%.3d', [ST.wHour, ST.wMinute, ST.wSecond, ST.wMilliseconds]);
end;

procedure TMainForm.ClearAllClick(Sender: TObject);
var
  J: Integer;
  MainChecks: TCheckListBox;

begin
  MainChecks := TCheckListBox(Pointer(TComponent(Sender).Tag));
  for J := 0 to MainChecks.Items.Count - 1 do
    MainChecks.Checked[J] := False
  ;
end;

procedure TMainForm.SelectAllClick(Sender: TObject);
var
  J: Integer;
  MainChecks: TCheckListBox;

begin
  MainChecks := TCheckListBox(Pointer(TComponent(Sender).Tag));
  for J := 0 to MainChecks.Items.Count - 1 do
    MainChecks.Checked[J] := True
  ;
end;

procedure FilterPax(Pax: TStringList);
var
  FilterForm: TForm;
  I: Integer;
  TopPanel, BottomPanel: TPanel;
  MainChecks: TCheckListBox;

begin
  FilterForm := TForm.Create(MainForm);
  with FilterForm do begin
    Position := poScreenCenter;
    Height := MainForm.Height * 2;
  end;

  MainChecks := TCheckListBox.Create(FilterForm);
  with MainChecks do begin
    Align := alClient;
    Parent := FilterForm;
    Items.Text := Pax.Text;
    for I := 0 to Items.Count - 1 do Checked[I] := True;
  end;

  TopPanel := TPanel.Create(FilterForm);
  with TopPanel do begin
    Height := 25;
    Align := alTop;
    Parent := FilterForm;
  end;

  with TButton.Create(FilterForm) do begin
    Caption := 'Select All';
    Tag := Integer(Pointer(MainChecks));
    OnClick := MainForm.SelectAllClick;
    Align := alLeft;
    Parent := TopPanel;
  end;

  with TButton.Create(FilterForm) do begin
    Caption := 'Clear All';
    Tag := Integer(Pointer(MainChecks));
    OnClick := MainForm.ClearAllClick;
    Align := alLeft;
    Parent := TopPanel;
  end;

  BottomPanel := TPanel.Create(FilterForm);
  with BottomPanel do begin
    Height := 25;
    Align := alBottom;
    Parent := FilterForm;
  end;

  with TButton.Create(FilterForm) do begin
    Caption := 'OK';
    ModalResult := mrOk;
    Align := alRight;
    Parent := BottomPanel;
  end;

  with TButton.Create(FilterForm) do begin
    Caption := 'Cancel';
    ModalResult := mrCancel;
    Align := alRight;
    Left := 10000;
    Parent := BottomPanel;
  end;

  Beep;
  if FilterForm.ShowModal <> mrOk then begin
    Pax.Clear;
    Exit;
  end;

  Pax.Text := MainChecks.Items.Text;

  for I := MainChecks.Items.Count - 1 downto 0 do
    if not MainChecks.Checked[I] then
      Pax.Delete(I)
  ;

  FilterForm.Free;

  Application.ProcessMessages;
end;

procedure TMainForm.GoButtonClick(Sender: TObject);
var
  CDPath, ThisPak, Path, oPath, CoNick, OutPath, MainINF, InPath, TarPath: String;
  MiniAction: TMiniAction;
  bMiniWarn, bPCI: Boolean;
  I, J, K: Integer;
  MI: TMemINIFile;
  Pax, Dirs, Files, FFiles, Content, OFiles, Dirs2Dirs: TStringList;
  Installer, Drivers, PCIFind, PackStr: String;
  nDrivers: Integer;
  Ages: TList;
  Completed: Integer;
  NuP: TTintProgress;
  LastTC, ThisTC: LongWord;

    {
    procedure LoadDrivers(BranchType: Char; Ext: String);
    var
      I: Integer;

    begin
      Files := DumpFileNamesToStringList(CDPath + 'Drivers\*.' + Ext);
      for I := 0 to Files.Count - 1 do begin
        FFiles.Add(BranchType + Files[I]);
        Content.Add(DumpFileToString(CDPath + 'Drivers\' + Files[I]));
        OFiles.Add('Z:\Drivers\' + Files[I]);
        Ages.Add(Pointer(FileAge(CDPath + 'Drivers\' + Files[I])));
      end;
      Files.Free;
    end;
    }

    procedure LoadFile(BranchType: Byte; FromPath: String; ToPath: String = '');
    begin
      if Length(ToPath) = 0 then ToPath := ExtractFileName(FromPath);

      FFiles.Add(Char(BranchType) + ToPath);
      Content.Add(DumpFileToString(CDPath + FromPath));
      OFiles.Add('Z:\' + FromPath);
      Ages.Add(Pointer(FileAge(CDPath + FromPath)));
    end;

    procedure LoadDrivers(BranchType: Byte; Ext: String; FromSubPath: String = ''; ToSubPath: String = '');
    var
      I: Integer;

    begin
      FromSubPath := 'Drivers\' + FromSubPath;
      Files := DumpFileNamesToStringList(CDPath + FromSubPath + '*.' + Ext);
      for I := 0 to Files.Count - 1 do
        LoadFile(BranchType, FromSubPath + Files[I], ToSubPath + Files[I])
      ;
      {
      for I := 0 to Files.Count - 1 do begin
        FFiles.Add(BranchType + ToSubPath + Files[I]);
        Content.Add(DumpFileToString(CDPath + FromSubPath + Files[I]));
        OFiles.Add('Z:\' + FromSubPath + Files[I]);
        Ages.Add(Pointer(FileAge(CDPath + FromSubPath + Files[I])));
      end;
      }
      Files.Free;
    end;

begin
  Completed := 0;
  NuP := nil;
  try
    GoButton.Visible := False;
    MessList.Items.Clear;
    NuP := TTintProgress.Create(Self);
    with NuP do begin
      Parent := GoPanel;
      Align := alClient;
      Tint := clBlue;
    end;

    MessList.Items.Add('Process begun at ' + TimeStr + '...');
    MessList.Update;

    CDPath := IncludeTrailingBackslash(PathEdit.Text);
    if not DirectoryExists(CDPath) then begin
      MessList.Items.Add('Fatal: CD path does not exist.');
      Exit;
    end;
    PathEdit.Text := CDPath;

    CoNick := CoNickEdit.Text;

    MainINF := INFEdit.Text;
    if not FileExists(CDPath + MainINF) then begin
      MessList.Items.Add('Fatal: Main .INF not found.');
      Exit;
    end;

    InPath := InstSrcEdit.Text;
    if not FileExists(InPath) then begin
      MessList.Items.Add('Fatal: installer source not found.');
      Exit;
    end;

    if not PushCheck.Checked then begin
      OutPath := IncludeTrailingBackslash(OutEdit.Text);
      if not DirectoryExists(OutPath) then begin
        MessList.Items.Add('Fatal: Output path does not exist.');
        Exit;
      end;
      OutEdit.Text := OutPath;
    end;

    I := MiniPackRadio.ItemIndex;
    MiniAction := TMiniAction(I div 2);
    bMiniWarn := ((I and 1) <> 0) or (MiniAction = maFail);



    Installer := DumpFileToString(InPath);
    I := Pos(SizerStr, Installer);
    if I = 0 then begin
      MessList.Items.Add('Fatal: Installer does not contain sizer string "' + SizerStr + '".');
      Exit;
    end;
    J := Pos(SizerStr, Copy(Installer, I + 1, MAXINT));
    if J <> 0 then begin
      MessList.Items.Add('Fatal: Installer contains multiple copies of sizer string "' + SizerStr + '".');
      Exit;
    end;
    PLongWord(PChar(@Installer[I + 4]))^ := Length(Installer);

    Pax := TStringList.Create;
    Dirs := TStringList.Create;
    //Files := TStringList.Create; //Dynamically reallocated by DumpAllSubFileNamesToStringList
    Ages := TList.Create;
    FFiles := TStringList.Create;
    Content := TStringList.Create;
    OFiles := TStringList.Create;
    Dirs2Dirs := TStringList.Create;

    LoadDrivers($4, 'sys');
    for I := 0 to OFiles.Count - 1 do
      if (UpperCase(OFiles[I]) = 'Z:\DRIVERS\AIO16.SYS')
      or (UpperCase(OFiles[I]) = 'Z:\DRIVERS\W2KPDD.SYS')
      //or (UpperCase(OFiles[I]) = 'Z:\DRIVERS\AIOUSB.SYS')
      //or (UpperCase(OFiles[I]) = 'Z:\DRIVERS\PCI8S.SYS')
      then
        FFiles[I] := #$24 + Copy(FFiles[I], 2, MAXINT)
    ;
    LoadDrivers($2, 'dll');
    LoadDrivers($D, 'dll', 'DLL.64\');
    LoadDrivers($3, 'vxd');
    LoadDrivers($28, 'sys', 'NT32\');
    LoadDrivers($29, 'sys', 'NT64\');
    LoadDrivers($2C, 'dll', 'NT32\');
    LoadDrivers($2D, 'dll', 'NT64\');
    LoadDrivers($31, 'spt', 'SPT\', 'System32\AIOSPT\');
    LoadFile($1, 'Drivers\Uninstall.exe', CoNick + 'Uninstall.exe');
    {
    FFiles.Add(#$1 + CoNick + 'Uninstall.exe');
    Content.Add(DumpFileToString(CDPath + 'Drivers\Uninstall.exe'));
    OFiles.Add('Z:\Drivers\Uninstall.exe');
    Ages.Add(Pointer(FileAge(CDPath + 'Drivers\Uninstall.exe')));
    }

    LoadFile($25, MainINF);
    LoadFile($25, 'AIOCyUSB.inf');
    LoadFile($25, 'Ports.inf');
    LoadFile($20, ChangeFileExt(MainINF, '.cat'));

    Drivers := '';
    for I := 0 to FFiles.Count - 1 do
      Drivers := Drivers + (StrToLStr(FFiles[I]) + StrToLStr(Content[I]) + StrToLStr(OFiles[I]) + IntAsStr(Integer(Ages[I])))
    ;
    nDrivers := FFiles.Count;
    FFiles.Clear;
    Content.Clear;
    OFiles.Clear;
    Ages.Clear;

    PCIFind := StrToLStr(#$1'PCIFind.exe') + StrToLStr(DumpFileToString(CDPath + 'Drivers\PCIFind.exe')) + StrToLStr('Z:\Drivers\PCIFind.exe') + IntAsStr(FileAge(CDPath + 'Drivers\PCIFind.exe'));

    MI := TMemINIFile.Create(CDPath + 'disks\disks.ini');
    MI.ReadSectionValues('Packages', Pax);
    MI.ReadSectionValues('files', Dirs2Dirs);
    MI.Free;

    FilterPax(Pax);

    for I := Dirs2Dirs.Count - 1 downto 0 do
      if Dirs2Dirs[I][Length(Dirs2Dirs[I])] = '=' then
        Dirs2Dirs.Delete(I)
    ;
    Dirs2Dirs.Insert(0, '');

    MessList.Items.Add('Overhead complete at ' + TimeStr + ', building ' + IntToStr(Pax.Count) + ' packages...');
    MessList.Update;

    NuP.Max := Pax.Count;
    for I := 0 to Pax.Count - 1 do begin
      NuP.Position := I;
      NuP.Paint;

      ThisPak := Pax.Names[I];
      if not DirectoryExists(CDPath + 'disks\' + ThisPak + '\') then begin
        MessList.Items.Add('Package ' + ThisPak + ' lacks a default directory.');
        MessList.Update;
      end;
      Dirs2Dirs[0] := 'default=' + ThisPak + ',TOOLS,TOOLS.win';
      Dirs.Text := StringReplaceC(RValue(Pax[I]), ',', #13);
      for J := 0 to Dirs.Count - 1 do begin
        if Dirs[J] = 'PTLDrv' then begin
          //Load the PTL drivers in "cyst-only" form.
          LoadFile($20, 'Drivers\PTL\PTLCom.cat');
          LoadFile($20, 'Drivers\PTL\PTLPort.dll');
          LoadFile($20, 'Drivers\PTL\PTLPort64.dll');
          LoadFile($20, 'Drivers\PTL\PTLEnum.inf');
          LoadFile($20, 'Drivers\PTL\PTLSer.inf');
          LoadFile($20, 'Drivers\PTL\PTLSer.sys');
          LoadFile($20, 'Drivers\PTL\PTLSer64.sys');
        end;
        if Dirs[J] = 'FTDIDrv' then begin
          //Load the FTDI drivers in "cyst-only" form.
          LoadFile($20, 'Drivers\FT\ftdibus.cat');
          LoadFile($20, 'Drivers\FT\ftdiport.cat');
          LoadFile($20, 'Drivers\FT\ftdibus.inf');
          LoadFile($20, 'Drivers\FT\ftdiport.inf');
          LoadFile($20, 'Drivers\FT\amd64\ftbusui.dll');
          LoadFile($20, 'Drivers\FT\amd64\ftcserco.dll');
          LoadFile($20, 'Drivers\FT\amd64\ftd2xx64.dll');
          LoadFile($20, 'Drivers\FT\amd64\ftlang.dll');
          LoadFile($20, 'Drivers\FT\amd64\ftserui2.dll');
          LoadFile($20, 'Drivers\FT\amd64\ftd2xx.lib');
          LoadFile($20, 'Drivers\FT\amd64\ftdibus.sys');
          LoadFile($20, 'Drivers\FT\amd64\ftser2k.sys');
          LoadFile($20, 'Drivers\FT\i386\ftbusui.dll');
          LoadFile($20, 'Drivers\FT\i386\ftcserco.dll');
          LoadFile($20, 'Drivers\FT\i386\ftd2xx.dll');
          LoadFile($20, 'Drivers\FT\i386\ftlang.dll');
          LoadFile($20, 'Drivers\FT\i386\ftserui2.dll');
          LoadFile($20, 'Drivers\FT\i386\ftd2xx.lib');
          LoadFile($20, 'Drivers\FT\i386\ftdibus.sys');
          LoadFile($20, 'Drivers\FT\i386\ftser2k.sys');
        end;
        Dirs[J] := StringReplaceC(Dirs2Dirs.Values[Dirs[J]], ',', #13);
      end;
      Dirs.Text := Dirs.Text; //Unfold multiline entries from Dirs2Dirs
      for J := Dirs.Count - 1 downto 0 do
        if Dirs[J] = '' then
          Dirs.Delete(J)
      ;

      bPCI := Dirs.IndexOf('tools.pci') <> -1;

      for J := 0 to Dirs.Count - 1 do begin
        Path := 'disks\' + Dirs[J] + '\';
        //oPath := 'Z:\' + Path;
        //Path := CDPath + Path;
        Files := DumpAllSubFileNamesToStringList(CDPath + Path);
        for K := 0 to Files.Count - 1 do begin
          LoadFile($10, Path + Files[K], Files[K]);
          {
          FFiles.Add(#$10 + Files[K]);
          Content.Add(DumpFileToString(Path + Files[K]));
          OFiles.Add(oPath + Files[K]);
          Ages.Add(Pointer(FileAge(Path + Files[K])));
          }
        end;
        Files.Free;
      end;

      if bPCI then
        PackStr := PCIFind
      else
        PackStr := ''
      ;
      LastTC := 0;
      for J := 0 to FFiles.Count - 1 do begin
        ThisTC := GetTickCount;
        if ThisTC <> LastTC then begin
          Application.ProcessMessages;
          LastTC := ThisTC;
        end;
        PackStr := PackStr + (StrToLStr(FFiles[J]) + StrToLStr(Content[J]) + StrToLStr(OFiles[J]) + IntAsStr(Integer(Ages[J])));
      end;
      PackStr := StrToIncLStr(StrToLStr(CoNick) + StrToLStr(ThisPak) + IntAsStr(nDrivers + Ord(bPCI) + FFiles.Count) + Drivers + PackStr);

      FFiles.Clear;
      Content.Clear;
      OFiles.Clear;
      Ages.Clear;

      if Length(PackStr) >= $40000000 then begin
        MessList.Items.Add('Package ' + ThisPak + ' was 1GB or larger.');
        MessList.Update;
        PackStr := '';
        Continue;
      end;

      if Length(PackStr) < $1000 then begin
        if bMiniWarn then begin
          MessList.Items.Add('Package ' + ThisPak + ' was under 4KB.');
          MessList.Update;
        end;
        case MiniAction of
          maPass: ;
          maPad: PackStr := PackStr + StringOfChar(#0, $1000 - Length(PackStr));
          maFail: begin
            PackStr := '';
            Continue;
          end;
        end;
      end;

      PackStr := Installer + PackStr;

      if PushCheck.Checked then begin
        TarPath := WebPackPath + ThisPak + ' Install.exe';
        DumpStringToFile(PackStr, TarPath);
        PackStr := '';
        TarPath := '"' + TarPath + '"';
        ShellExecute(0, nil, 'P:\FTPPush.exe', PChar(TarPath), nil, SW_SHOWNORMAL);
        TarPath := '';
      end
      else begin
        TarPath := OutPath + ThisPak + ' Install.exe';
        DumpStringToFile(PackStr, TarPath);
        PackStr := '';
      end;
      Inc(Completed);
    end;

    Pax.Free;
    Dirs.Free;
    //Files.Free;
    Ages.Free;
    FFiles.Free;
    Content.Free;
    OFiles.Free;
    Dirs2Dirs.Free;

  finally
    GoButton.Visible := True;

    MessList.Items.Add('Process ended at ' + TimeStr + '.');
    MessList.Items.Add(IntToStr(Completed) + ' packages built.');

    FreeAndNil(NuP);
  end;
end;

procedure TMainForm.FormCreate(Sender: TObject);
begin
  CoNickEdit.Align := alTop;
  PathEdit.Align := alTop;
end;

procedure TMainForm.PushCheckClick(Sender: TObject);
begin
  if PushCheck.Checked then begin
    OutEdit.ParentColor := True;
    OutEdit.ParentFont := True;
  end
  else begin
    OutEdit.Color := clWindow;
    OutEdit.Font.Color := clWindowText;
  end;
end;

end.
